#!/usr/bin/env python3

"""Utilities related to binary

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
from typing import List, Union
import numpy as np
# endregion 

# region Logic
def message2binary(message: Union[str, bytes, np.ndarray, int, np.uint8]) -> Union[str, List[str]]:
  """Utility function for converting strings into their bit representation
  ----------
  Parameters
  ----------
  message: str
    Message to be converted into bitlike format
  -------
  Returns
  -------
  str
    bit like representation of given string
  ------
  Raises
  ------
  TypeError
    when type of message is nor string, int like or bytelike object.
  """
  if type(message) == str:
      result = ''.join([format(ord(char), "08b") for char in message])
  elif type(message) in (bytes, np.ndarray):
      result = [format(char, "08b") for char in message]
  elif type(message) in (int, np.uint8):
      result = format(message, "08b")
  else:
      raise TypeError("Input type is not supported")
  return result
# endregion



# region Imports
from typing import Dict
import argparse
import logging

from .convert import decode, encode
from src.types.Modes import Modes

# endregion

# region CLI stuff

PARSER = argparse.ArgumentParser()
PARSER.add_argument("-m", "--mode", type=Modes,
                    choices=list(Modes), metavar="decode", required=True)
PARSER.add_argument("-p", "--payload", type=str, metavar="Test message")
PARSER.add_argument("-o", "--output", type=str, metavar="output.png")
PARSER.add_argument("-i", "--input", type=str,
                    metavar="inputFile.png", required=True)
LOGGER = logging.getLogger(__name__)
# endregion

def run():
  args: Dict[str, any] = PARSER.parse_args()
  logging.basicConfig(level=logging.DEBUG,
                      format=" %(name)s :: %(levelname)-8s :: %(message)s")
  if args.mode == Modes.ENCODE:
      if not args.payload:
          raise ValueError("Payload is required")
      elif not args.output:
          raise ValueError("output file is required")
      elif not args.output.endswith(".png"):
          raise ValueError("Output file has to be png")
      encode(args.input, args.payload, args.output)
  elif args.mode == Modes.DECODE:
      if not args.input.endswith(".png"):
          raise ValueError("Input file has to be png")
      decode(args.input)

if __name__ == "__main__":
  run()


#!/usr/bin/env python3

"""File holding utilities related to image processing.

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
import logging

from src.decorators.notNone import NotNone

# endregion

# region Constants 
LOGGER = logging.getLogger(__name__)
# endregion

# region Logic
@NotNone()
def verify_payload(payload: str, img: any) -> None:
  """
  ----------------
  Verifies payload
  ----------------

  Checks if input image is big enough to write given payload in

  ----------
  Parameters
  ----------
  payload: str
    payload to write into image
  img: Image
    image object representing image into which payload is to be written
  ------
  Raises
  ------
  ValueError:
    if payload is too big for given image
  """
  max_bytes = (img.shape[0] * img.shape[1] * 3) // 8
  LOGGER.debug(f"Maximum bytes that can be encoded {max_bytes}")
  if len(payload) > max_bytes:
    LOGGER.error(f"Payload is too big (Max bytes: {max_bytes})")
    raise ValueError("Payload is too big")
# endregion



#!/usr/bin/env python3

"""Type representing modes in which can be script run.

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
from enum import Enum
# endregion

# region Logic
class Modes(Enum):
  """
  ----------
  Modes enum
  ----------
  Enum for modes in which script can be run
  
  ----------
  Attributes
  ----------
  ENCODE : str
    encode mode
  DECODE: str
    decode mode
  """
  DECODE = "decode"
  ENCODE = "encode"

  def __str__(self) -> str:
      """String representation of mode
      -------
      Returns
      -------
      string: str
        string representation of given mode
      """
      return self.value
# endregion


#!/urs/bin/env python3

"""Type representing RGB color components with order in pixel

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
from enum import IntEnum
# endregion

class RGB(IntEnum):
  """
  --------
  RGB enum
  --------
  Enum that represents color components in each pixel.

  Attributes
  ----------
  RED: int
    index of red component in pixel
  GREEN: int
    index of green component
  BLUE: int
    index of blue component in pixel
  """
  RED = 0;
  GREEN = 1;
  BLUE = 2;


#!/usr/bin/env python3

"""Decorator for arguments which can not be null.

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
import logging
# endregion

# region Constants
LOGGER = logging.getLogger(__name__)
# endregion

# region Logic
def NotNone(*names):
  """
  -------
  NotNone
  -------
  
  Decorator that throws ValueError if any on non keyword arguments are None of keywords with given names are None
  
  ----------
  Parameters
  ----------
  names: List[str]
    names of keyword arguments to validate
  ------
  Raises
  ------
  ValueError:
    if any of non keyword arguments are None or any keyword argument with given name are None 
  """
  def _notNone(function):
    def __notNone(*args, **kwargs):
      for arg in args:
        try:
          not arg
        except ValueError:
          continue
        if not arg:
          LOGGER.error("Non keyword arguments can not be null")
          raise ValueError("Non keyword arguments can not be null")
      if names:
        for name in names:
          if not kwargs.get(name):
            LOGGER.error(f"Argument `{name}` can not be None")
            raise ValueError(f"Argument `{name}` can not be None")
      return function(*args, **kwargs)
    return __notNone
  return _notNone
# endregion



#!/urs/bin/env python3

"""Module storing all `types`
Types stored in this module are simple classes with minimal logic

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
from .Modes import Modes
from .ColorComponents import RGB
# endregion


#!/usr/bin/env/ python3

"""Module for utility functions.

----------
Since:
  1.0
----------
------------
Version:
  1.0
------------
---------------
Author:
  Karafra
---------------
"""

# region Imports
from .binary import message2binary
from .stegUtils import verify_payload
# endregion


#!/usr/bin/env python3

"""Convertors
File storing functions for encoding and decoding messages into/from images

--------
Version: 1.0
------------
----------
Since: 1.0
----------
---------------
Author: Karafra
---------------
"""

# region Imports
import logging

import cv2

from src.decorators.notNone import NotNone
from src.types.ColorComponents import RGB
from src.utilities.binary import message2binary
from src.utilities.stegUtils import verify_payload

# endregion 

# region Constants
LOGGER = logging.getLogger(__name__)
# endregion

# region Logic
@NotNone()
def encode(input_path: any, payload: str, output_path: str, terminator: str = "#####") -> None:
    """
    ------
    Encode
    ------

    Encodes string into given image and saves as output image.

    ----------
    Parameters
    ----------
    img: any
      image to which payload is to be hidden
    pyaload: str
      payload to hide
    output: str
      path to output file
    terminator: str (optional)
      terminator of message
    """
    # Check is payload can fit into image
    img = cv2.imread(input_path)
    verify_payload(payload, img)
    
    BINARY_DATA = message2binary(payload)
    LOGGER.debug(f"Binary payload: {BINARY_DATA}")
    
    # Append terminator and convert payload to list
    payload_finalized = list(message2binary(payload + terminator))
    for line in img:
        for pixel in line:
            bit_encoded = message2binary(pixel)
            for component in list(RGB):
              if not payload_finalized:
                cv2.imwrite(output_path, img)
                return LOGGER.info(f"Saved as {output_path}")
              bit = payload_finalized.pop(0)
              pixel[component.value] = int(
                bit_encoded[component.value][:-1] + bit,
                2
              )


@NotNone()
def decode(input_path: any, terminator: str = "#####") -> None:
  """
  -------------
  Decodes image
  -------------

  Decodes image and returns message hidden in it if any exists

  ----------
  Parameters
  ----------
  img: any
    image object to check for information
  teminator: str
    terminator of message
  """
  img = cv2.imread(input_path)
  retrieved_data = ""
  for line in img:
      for pixel in line:
          bit_encoded = message2binary(pixel)
          for component in list(RGB):
            retrieved_data += bit_encoded[component.value][-1]
  all_bytes = [
    retrieved_data[i: i+8] for i in range(0, len(retrieved_data), 8)
  ]
  decoded_data = ""
  for byte in all_bytes:
      # Convert from base 2
      decoded_data += chr(int(byte, 2))
      # Check if terminator was reached.
      if decoded_data[-5:] == terminator:
          break
  logging.info(f"Encoded data : {decoded_data[:-5]}")
# endregion
